"""Provides the preview applications for the textual CLI tool."""

from .borders import BorderApp
from .colors import ColorsApp
from .easing import EasingApp
from .keys import KeysApp

__all__ = ["BorderApp", "ColorsApp", "EasingApp", "KeysApp"]
